/*
 *    Copyright 2010 Jonathan Savage
 * 
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Xml.Serialization;
using System.IO;
using System.Globalization;

namespace ParseTCX
{
    class Program
    {
        const string version = "Version 1.2";
        const int roundTo = 4;
        static void Main(string[] args)
        {
            Console.Out.WriteLine(version);
            if (args.Length == 0)
            {
                LookForFiles();
            }
            else if (args.Length > 1)
            {
                Console.Out.WriteLine("Usage: ParseTCX [<file.tcx>] (too many parameters - did you forget to quote the filename?)");
            }
            else
            {
                ProcessFile(args[0], "", false);
            }

            Console.Out.WriteLine("Press return to contine");
            Console.In.ReadLine();
        }

        private static void LookForFiles()
        {
            string appdata = Environment.GetFolderPath(Environment.SpecialFolder.ApplicationData);

            //C:\Documents and Settings\a560507\Application Data\GARMIN\Devices\3807983990\History
            string devicePathXP = appdata + @"\GARMIN\Devices";
            if (!Directory.Exists(devicePathXP))
            {
                Console.Out.WriteLine("Could not find " + devicePathXP);
                return;
            }
            DirectoryInfo devices = new DirectoryInfo(devicePathXP);
            LookForFiles(devices);
        }

        private static void LookForFiles(DirectoryInfo start)
        {
            foreach (DirectoryInfo dir in start.GetDirectories())
            {
                LookForFiles(dir);
            }
            foreach (FileInfo file in start.GetFiles("*.tcx"))
            {
                ProcessFile(file.FullName, file.Name, true);
            }
        }


        private static void ProcessFile(string filename, string shortname, bool summary)
        {
            TrainingCenterDatabase_t tcx;
            CultureInfo cultures = new CultureInfo("en-US");

            XmlSerializer mySerializer = new XmlSerializer(typeof(TrainingCenterDatabase_t));
            //FileStream myFileStream = new FileStream(@"C:\Documents and Settings\a560507\My Documents\A Temp\12_3_2010 7_02_56 AM_history.tcx", FileMode.Open);
            //FileStream myFileStream = new FileStream(@"C:\Documents and Settings\a560507\My Documents\A Temp\12_1_2010 7_00_32 AM_history.tcx", FileMode.Open);
            if (!File.Exists(filename))
            {
                Console.Out.WriteLine("File " + filename + " not found");
                return;
            }
            try
            {
                FileStream myFileStream = new FileStream(filename, FileMode.Open);
                tcx = (TrainingCenterDatabase_t)mySerializer.Deserialize(myFileStream);
            }
            catch (Exception e)
            {
                Console.Out.WriteLine("Oops - encountered an error handling " + filename);
                Console.Out.WriteLine("Exception is " + e);
                return;
            }

            double sumscale = 0.0d;
            int numpoints = 0;
            double perlap_sumscale = 0.0d;
            int perlap_numpoints = 0;
            List<double> scales = new List<double>();
            if (tcx.Activities == null || tcx.Activities.Activity == null)
            {
                if (!summary)
                    Console.Out.WriteLine("No activity found in " + filename);
                return;
            }
            List<double> lapscale = new List<double>();
            foreach (Activity_t activity in tcx.Activities.Activity)
            {
                if (activity.Lap != null)
                {
                    foreach (ActivityLap_t lap in activity.Lap)
                    {
                        DateTime lasttime = DateTime.Now;
                        double lastdistance = 0.0d;
                        bool valid = false;
                        double lapsumscale = 0.0d;
                        int lapnumpoints = 0;
                        if (lap.Track != null)
                        {
                            foreach (Trackpoint_t point in lap.Track)
                            {
                                if (point.Extensions != null &&
                                    point.Extensions.Any != null && 
                                    point.Extensions.Any.Length > 0 &&
                                    point.Extensions.Any[0].ChildNodes != null &&
                                    point.Extensions.Any[0].ChildNodes.Count > 1 &&
                                    point.Extensions.Any[0].ChildNodes[0].LocalName == "Speed" &&
                                    point.Extensions.Any[0].ChildNodes[1].LocalName == "RunCadence")
                                {
                                    //got a good one
                                    if (valid)
                                    {
                                        double seconds = (double)point.Time.Subtract(lasttime).TotalSeconds;
                                        double currentdistance = (double)point.DistanceMeters;
                                        double distance = currentdistance - lastdistance;
                                        if (seconds > 0 && distance > 0)
                                        {
                                            double gpsSpeed = distance / seconds;
                                            double footpodspeedMPS = double.Parse(point.Extensions.Any[0].FirstChild.ChildNodes[0].Value, cultures); //Meters per second!
                                            if (footpodspeedMPS > 0.0d)
                                            {
                                                double footpoddistance = footpodspeedMPS * seconds;
                                                double scale = distance / footpoddistance;
                                                scales.Add(scale);
                                                //Console.Out.WriteLine("GPS Speed (m/s) " + double.Round(gpsSpeed, 2) + ", Foot Pod Speed (m/s) " + double.Round(footpodspeedMPS, 2) + ", footpod distance " + double.Round(footpoddistance, 2) + ", gps distance " + double.Round(distance, 2) + ", seconds " + seconds + ", scale " + double.Round(scale, 4));
                                                sumscale += scale;
                                                numpoints++;
                                                lapsumscale += scale;
                                                lapnumpoints++;
                                            }
                                        }
                                    }
                                    lasttime = point.Time;
                                    lastdistance = (double)point.DistanceMeters;
                                    valid = true;
                                }
                            }

                            if (!summary && lapnumpoints > 0)
                            {
                                lapscale.Add(lapsumscale / lapnumpoints);
                                perlap_sumscale += (lapsumscale / lapnumpoints);
                                perlap_numpoints++;
                            }
                        }
                    }
                }
            }
            double stddev = GetStdev(scales);
            string rating = GetRating(stddev);
            if (summary)
            {
                if (numpoints == 0)
                {
                    Console.Out.WriteLine(shortname + " No usable data was found in ");
                }
                else
                {
                    Console.Out.WriteLine(shortname + " calibration factor adjustment " + Math.Round(sumscale / numpoints, roundTo) +
                        ", Consistency is " + rating + " (stddev " + Math.Round(stddev, roundTo) + ")");
                }
            }
            else
            {
                if (numpoints == 0)
                {
                    Console.Out.WriteLine("Sorry, no usable data was found");
                }
                else
                {
                    if (lapscale.Count > 1)
                    {
                        Console.Out.WriteLine("Per Lap Scaling");
                        Console.Out.WriteLine("===============");
                        foreach (double s in lapscale)
                        {
                            Console.Out.WriteLine("lap average scale " + Math.Round(s, roundTo));
                        }
                        Console.Out.WriteLine();
                    }
                    Console.Out.WriteLine("Details");
                    Console.Out.WriteLine("=======");
                    Console.Out.WriteLine("Average scale              " + Math.Round(sumscale / numpoints, roundTo) + ", " + Math.Round(sumscale / numpoints * 100.0, roundTo) + "%");
                    Console.Out.WriteLine("Consistency                " + rating);
                    Console.Out.WriteLine("  Standard Deviation scale " + Math.Round(stddev, roundTo) + ", " + Math.Round(stddev * 100.0, roundTo) + "%");
                    Console.Out.WriteLine("  Scale+stddev             " + (Math.Round(sumscale / numpoints, roundTo) + Math.Round(GetStdev(scales), roundTo)) + ", " + (Math.Round(sumscale / numpoints, roundTo) + Math.Round(GetStdev(scales), roundTo)) * 100.0 + "%");
                    Console.Out.WriteLine("  Scale-stddev             " + (Math.Round(sumscale / numpoints, roundTo) - Math.Round(GetStdev(scales), roundTo)) + ", " + (Math.Round(sumscale / numpoints, roundTo) - Math.Round(GetStdev(scales), roundTo)) * 100.0 + "%");
                    if (perlap_numpoints > 1)
                        Console.Out.WriteLine("Average scale (per lap)  " + Math.Round(perlap_sumscale / perlap_numpoints, roundTo) + ", " + Math.Round(perlap_sumscale / perlap_numpoints * 100.0, roundTo) + "%");
                    Console.Out.WriteLine("");
                    Console.Out.WriteLine("Calibration Instructions");
                    Console.Out.WriteLine("========================");
                    Console.Out.WriteLine("");
                    Console.Out.WriteLine("Multiply your current calibration factor by " + Math.Round(sumscale / numpoints, roundTo));
                    Console.Out.WriteLine();
                }
            }
        }

        private static string GetRating(double stddev)
        {
            string rating;
            if (stddev < 0.02)
                rating = "very good";
            else if (stddev < 0.05)
                rating = "good";
            else if (stddev < 0.075)
                rating = "okay";
            else if (stddev < 0.1)
                rating = "poor";
            else if (stddev < 0.2)
                rating = "very poor";
            else
                rating = "appalling";
            return rating;
        }


        static private double GetAverage(List<double> list)
        {
            double retval = 0;
            foreach (double d in list)
            {
                retval += d;
            }
            return retval / list.Count;
        }


        static public double GetVariance(List<double> list)
        {
            int len = list.Count;
            // Get average
            double avg = GetAverage(list);

            double sum = 0;
            foreach (double d in list)
            {
                sum += Math.Pow((d - avg), 2);
            }
            return sum / len;
        }
        /// <summary>
        /// Get standard deviation
        /// </summary>
        static public double GetStdev(List<double> list)
        {
            return Math.Sqrt(GetVariance(list));
        } 

    }
}
